use std::cell::RefCell;
use std::collections::HashMap;
use std::os::unix::prelude::OwnedFd;
use std::rc::Rc;

mod msg;
// mod crypto;
mod helper;
mod io;
pub mod service;
// pub mod fleck_core;

pub mod prelude {
    pub use crate::helper::{AsAny, IntoWeak};
    pub use crate::msg::{Message,MessageParams};
    pub use crate::service::{
        ChannelSpec, EventRoot, Service,
    };
    // pub use crate::fleck_core;
}

use prelude::*;

pub struct Fleck {
    io: RefCell<io::IO>,
    fd_services: RefCell<HashMap<mio::Token, Rc<dyn io::AbstractFdService>>>,
    services: RefCell<service::ServiceStack>,
}

impl Fleck {
    pub fn new() -> Rc<Self> {
        let res = Rc::new(Self {
            io: RefCell::new(io::IO::new()),
            fd_services: Default::default(),
            services: Default::default(),
        });

        res.register_core_services();

        res
    }

    fn register_core_services(&self) {
        // initial outgoing/incoming/minor/major channels
        // self.services.borrow_mut().create_io_channels();

        /*
        // Node registration
        self.add_service::<fleck_core::Nodes>();
        // Local node discovery
        self.add_service::<fleck_core::lowlevel::LocalDiscovery>();
        // Actually sending packets
        self.add_service::<fleck_core::lowlevel::SendPacket>();
        // Parsing incoming packets
        self.add_service::<msg::ParsePacket>();
        // Signing packets
        self.add_service::<crypto::SignPacket>();
        */
    }

    pub fn run(&self) {
        let io = io::IO::new();

        // TODO: insert/build global socket, minor, major timerfds

        io.run(self);
    }
}

struct SockIncomingTag {}
struct SockOutgoingTag {}

impl Fleck {
    pub fn bind_udp_socket(&self, port: u16, multicast: Option<std::net::Ipv4Addr>) {
        let socket = mio::net::UdpSocket::bind(
            std::net::SocketAddrV4::new(std::net::Ipv4Addr::UNSPECIFIED, port).into(),
        )
        .expect("couldn't listen on UDP port?");

        if let Some(mcast) = multicast {
            socket 
                .join_multicast_v4(&mcast, &std::net::Ipv4Addr::UNSPECIFIED)
                .expect("couldn't join multicast group?");
        }

        self.io.borrow_mut().add_udp_service::<(SockIncomingTag, (std::net::SocketAddr, Vec<u8>)),(SockOutgoingTag, (std::net::SocketAddr, Vec<u8>)),_>(socket);
    }

    /// Add a UDP socket to the Fleck polling loop.
    pub fn add_udp_socket(&self, socket: mio::net::UdpSocket) {
        
    }
    /// Add a file descriptor (stream socket, pipe, etc) to the Fleck polling loop.
    pub fn add_raw_fd(&self, socket: OwnedFd) {
        //self.io.borrow_mut().add_service(io::
    }

    /// Add a service to the Fleck instance. Must not be invoked from a service context.
    pub fn add_service<S: Default + 'static>(&self)
    where
        Rc<S>: Service,
    {
        self.services.borrow_mut().give_service::<S>();
    }
    pub fn services(&self) -> std::cell::Ref<service::ServiceStack> {
        self.services.borrow()
    }
    pub fn with_service<S: 'static, R, F: FnOnce(&S) -> R>(&self, f: F) -> R
    where
        Rc<S>: Service,
    {
        self.services.borrow().with_service(f)
    }
    pub fn queue<CS: service::ChannelSpec>(&self, data: CS::Data) {
        self.services
            .borrow()
            .event_root()
            .channel::<CS>()
            .queue(data);
    }
}

/*impl io::IOFeedback for Fleck {
    fn packet(&self, packet: io::Packet) {
        self.services.borrow().process_incoming(self, packet);
    }

    fn minor_tick(&self) {
        self.services.borrow().process_minor_tick(self);
    }

    fn major_tick(&self) {
        self.services.borrow().process_major_tick(self);
    }
}*/
